# Theorem-like counters, extended, separately for each type of theorem

Format 111
Counter criterion
	GuiName Criterion
	PrettyFormat "Criterion|Criteria ##"
End
Counter algorithm
	GuiName Algorithm
	LaTeXName lyxalgorithm
	PrettyFormat "Algorithm|Algorithms ##"
End
Counter axiom
	GuiName Axiom
	LaTeXName ax
	PrettyFormat "Axiom|Axioms ##"
End
Counter condition
	GuiName Condition
	PrettyFormat "Condition|Conditions ##"
End
Counter note
	GuiName Note
	PrettyFormat "Note|Notes ##"
End
Counter notation
	GuiName Notation
	PrettyFormat "Notation|Notations ##"
End
Counter summary
	GuiName Summary
	PrettyFormat "Summary|Summaries ##"
End
Counter conclusion
	GuiName Conclusion
	PrettyFormat "Conclusion|Conclusions ##"
End
Counter assumption
	GuiName Assumption
	PrettyFormat "Assumption|Assumptions ##"
End
Counter question
	GuiName Question
	PrettyFormat "Question|Questions ##"
End
