# This file is generated by generate_incoding_info.py from lib/languages file.
# Do not change this file directly.

lang = {
    "afrikaans": ["afrikaans", "Afrikaans", "false", "iso8859-1", "af_ZA", ""],
    "american": ["american", "American", "false", "iso8859-1", "en_US", ""],
    "arabic": ["arabic", "Arabic", "true", "iso8859-6", "ar_SA", ""],
    "austrian": ["austrian", "Austrian", "false", "iso8859-1", "de_AT", ""],
    "bahasa": ["bahasa", "Bahasa", "false", "iso8859-1", "in_ID", ""],
    "basque": ["basque", "Basque", "false", "iso8859-1", "eu_ES", ""],
    "belarusian": ["belarusian", "Belarusian", "false", "cp1251", "be_BY", ""],
    "brazil": ["brazil", "Portuguese (Brazil)", "false", "iso8859-1", "pt_BR", ""],
    "breton": ["breton", "Breton", "false", "iso8859-1", "br_FR", ""],
    "british": ["british", "British", "false", "iso8859-1", "en_GB", ""],
    "bulgarian": ["bulgarian", "Bulgarian", "false", "cp1251", "bg_BG", ""],
    "canadian": ["canadian", "Canadian", "false", "iso8859-1", "en_CA", ""],
    "canadien": ["canadien", "French Canadian", "false", "iso8859-1", "fr_CA", ""],
    "catalan": ["catalan", "Catalan", "false", "iso8859-1", "ca_ES", ""],
    "croatian": ["croatian", "Croatian", "false", "iso8859-2", "hr_HR", ""],
    "czech": ["czech", "Czech", "false", "iso8859-2", "cs_CZ", ""],
    "danish": ["danish", "Danish", "false", "iso8859-1", "da_DK", ""],
    "default": ["default", "default", "false", "iso8859-1", "C", ""],
    "dutch": ["dutch", "Dutch", "false", "iso8859-1", "nl_NL", ""],
    "english": ["english", "English", "false", "iso8859-1", "en_US", ""],
    "esperanto": ["esperanto", "Esperanto", "false", "iso8859-3", "eo", ""],
    "estonian": ["estonian", "Estonian", "false", "iso8859-1", "et_EE", ""],
    "finnish": ["finnish", "Finnish", "false", "iso8859-1", "fi_FI", ""],
    "french": [
        "french",
        "French",
        "false",
        "iso8859-1",
        "fr_FR",
        "\\addto\\extrasfrench{\\providecommand{\\og}{\\leavevmode\\flqq~}\\providecommand{\\fg}{\\ifdim\\lastskip>\\z@\\unskip\\fi~\\frqq}}",
    ],
    "frenchb": [
        "french",
        "French",
        "false",
        "iso8859-1",
        "fr_FR",
        "",
    ],  # for compatibility reasons
    "galician": ["galician", "Galician", "false", "iso8859-1", "gl_ES", ""],
    "german": ["german", "German", "false", "iso8859-1", "de_DE", ""],
    "greek": ["greek", "Greek", "false", "iso8859-7", "el_GR", ""],
    "hebrew": ["hebrew", "Hebrew", "true", "cp1255", "he_IL", ""],
    "icelandic": ["icelandic", "Icelandic", "false", "iso8859-1", "is_IS", ""],
    "irish": ["irish", "Irish", "false", "iso8859-1", "ga_IE", ""],
    "italian": ["italian", "Italian", "false", "iso8859-1", "it_IT", ""],
    "kazakh": ["kazakh", "Kazakh", "false", "pt154", "kk_KZ", ""],
    "latvian": ["latvian", "Latvian", "false", "iso8859-13", "lv_LV", ""],
    "lithuanian": ["lithuanian", "Lithuanian", "false", "iso8859-13", "lt_LT", ""],
    "magyar": ["magyar", "Magyar", "false", "iso8859-2", "hu_HU", ""],
    "naustrian": [
        "naustrian",
        "Austrian (new spelling)",
        "false",
        "iso8859-1",
        "de_AT",
        "",
    ],
    "ngerman": ["ngerman", "German (new spelling)", "false", "iso8859-1", "de_DE", ""],
    "norsk": ["norsk", "Norsk", "false", "iso8859-1", "no_NO", ""],
    "nynorsk": ["nynorsk", "Nynorsk", "false", "iso8859-1", "nn_NO", ""],
    "polish": ["polish", "Polish", "false", "iso8859-2", "pl_PL", ""],
    "portuges": ["portuges", "Portugese", "false", "iso8859-1", "pt_PT", ""],
    "romanian": ["romanian", "Romanian", "false", "iso8859-2", "ro_RO", ""],
    "russian": ["russian", "Russian", "false", "koi8-r", "ru_RU", ""],
    "scottish": ["scottish", "Scottish", "false", "iso8859-1", "gd_GB", ""],
    "serbian": ["croatian", "Serbian", "false", "iso8859-5", "sr_HR", ""],
    "serbocroatian": ["croatian", "Serbo-Croatian", "false", "iso8859-2", "sh_HR", ""],
    "slovak": ["slovak", "Slovak", "false", "iso8859-2", "sk_SK", ""],
    "slovene": ["slovene", "Slovene", "false", "iso8859-2", "sl_SI", ""],
    "spanish": [
        "spanish",
        "Spanish",
        "false",
        "iso8859-1",
        "es_ES",
        "\\deactivatetilden",
    ],
    "swedish": ["swedish", "Swedish", "false", "iso8859-1", "sv_SE", ""],
    "thai": ["thai", "Thai", "false", "tis620-0", "th_TH", "\\usepackage{thswitch}"],
    "turkish": ["turkish", "Turkish", "false", "iso8859-9", "tr_TR", ""],
    "ukrainian": ["ukrainian", "Ukrainian", "false", "koi8-u", "uk_UA", ""],
    "welsh": ["welsh", "Welsh", "false", "iso8859-1", "cy_GB", ""],
}
