# Theorem-like counters, separately for each type of theorem

Format 111
Counter theorem
	GuiName Theorem
	LaTeXName thm
	PrettyFormat "Theorem|Theorems ##"
End
Counter corollary
	GuiName Corollary
	LaTeXName cor
	PrettyFormat "Corollary|Corollaries ##"
End
Counter lemma
	GuiName Lemma
	LaTeXName lem
	PrettyFormat "Lemma|Lemmas ##"
End
Counter proposition
	GuiName Proposition
	LaTeXName prop
	PrettyFormat "Proposition|Propositions ##"
End
Counter conjecture
	GuiName Conjecture
	PrettyFormat "Conjecture|Conjectures ##"
End
Counter fact
	GuiName Fact
	PrettyFormat "Fact|Facts ##"
End
Counter definition
	GuiName Definition
	LaTeXName defn
	PrettyFormat "Definition|Definitions ##"
End
Counter example
	GuiName Example
	PrettyFormat "Example|Examples ##"
End
Counter problem
	GuiName Problem
	PrettyFormat "Problem|Problems ##"
End
Counter exercise
	GuiName Exercise
	LaTeXName xca
	PrettyFormat "Exercise|Exercises ##"
End
Counter solution
	GuiName Solution
	LaTeXName sol
	PrettyFormat "Solution|Solutions ##"
End
Counter remark
	GuiName Remark
	LaTeXName rem
	PrettyFormat "Remark|Remarks ##"
End
Counter claim
	GuiName Claim
	PrettyFormat "Claim|Claims ##"
End
